@@ -1,7 +1,7 @@
-/* imcd.c */
+/* imagecodecs/imcd.c */
 
 /*
-Copyright (c) 2008-2023, Christoph Gohlke.
+Copyright (c) 2008-2025, Christoph Gohlke.
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
@@ -69,7 +69,7 @@ POSSIBILITY OF SUCH DAMAGE.
 /* Inplace swap bytes. */
 void imcd_swapbytes(
     void* src,
-    const ssize_t srcsize,
+    const ssize_t srcsize,  /* number items, not bytes */
     const ssize_t itemsize)
 {
     uint16_t* d16;
@@ -601,6 +601,10 @@ ssize_t imcd_packbits_decode_size(
     }
     while (srcptr < srcend) {
         n = (ssize_t)(*srcptr++) + 1;
+        if ((n == 0) && (srcptr == srcend)) {
+            /* DICOM pad byte */
+            break;
+        }
         if (n < 129) {
             /* literal */
             srcptr += n;
@@ -621,7 +625,8 @@ ssize_t imcd_packbits_decode(
     const uint8_t* src,
     const ssize_t srcsize,
     uint8_t* dst,
-    const ssize_t dstsize)
+    const ssize_t dstsize,
+    const ssize_t dststride)
 {
     uint8_t* srcptr = (uint8_t*)src;
     uint8_t* dstptr = dst;
@@ -637,16 +642,21 @@ ssize_t imcd_packbits_decode(
 
     while (srcptr < srcend) {
         n = (ssize_t)(*srcptr++) + 1;
+        if ((n == 1) && (srcptr == srcend)) {
+            /* DICOM pad byte */
+            break;
+        }
         if (n < 129) {
             /* literal */
             if (srcptr + n > srcend) {
                 return IMCD_INPUT_CORRUPT;
             }
-            if (dstptr + n > dstend) {
+            if (dstptr + (n - 1) * dststride >= dstend) {
                 return IMCD_OUTPUT_TOO_SMALL;
             }
             while (n--) {
-                *dstptr++ = *srcptr++;
+                *dstptr = *srcptr++;
+                dstptr += dststride;
             }
         }
         else if (n > 129) {
@@ -655,17 +665,18 @@ ssize_t imcd_packbits_decode(
             if (srcptr >= srcend) {
                 return IMCD_INPUT_CORRUPT;
             }
-            if (dstptr + n > dstend) {
+            if (dstptr + (n - 1) * dststride >= dstend) {
                 return IMCD_OUTPUT_TOO_SMALL;
             }
             e = *srcptr++;
             while (n--) {
-                *dstptr++ = e;
+                *dstptr = e;
+                dstptr += dststride;
             }
         }
         /* else if (n == 129) {NOP} */
     }
-    return (ssize_t)(dstptr - dst);
+    return ((ssize_t)(dstptr - dst)) / dststride;
 }
 
 /* Return maximum length of PackBits compressed sequence. */
@@ -680,12 +691,10 @@ inline uint8_t* _packbits_next_replicate(
     uint8_t* srcptr,
     const uint8_t* srcend)
 {
-    uint8_t value = *srcptr;
+    uint8_t value = (srcptr < srcend) ? *srcptr : 0;
 
-    while (++srcptr < srcend)
-    {
-        if (value == *srcptr)
-        {
+    while (++srcptr < srcend) {
+        if (value == *srcptr) {
             return srcptr - 1;
         }
         value = *srcptr;
@@ -700,7 +709,7 @@ inline ssize_t _packbits_replicate_length(
     const uint8_t* srcend)
 {
     uint8_t* srcptr = (uint8_t*)src;
-    const uint8_t value = *srcptr;
+    const uint8_t value = (srcptr < srcend) ? *srcptr : 0;
 
     while ((++srcptr < srcend) && (value == *srcptr)) {;}
     return (ssize_t)(srcptr - src);
@@ -733,8 +742,7 @@ ssize_t imcd_packbits_encode(
     while (srcptr < srcend)
     {
         dupptr = _packbits_next_replicate(srcptr, srcend);
-        if (srcptr == dupptr)
-        {
+        if (srcptr == dupptr) {
             /* replicate */
             replicate = _packbits_replicate_length(dupptr, srcend);
             replicate = MIN_(128, replicate);
@@ -754,14 +762,13 @@ ssize_t imcd_packbits_encode(
             literal = srcend - srcptr;
         }
         else {
+            /* try skip next replicate run < 3 */
             replicate = _packbits_replicate_length(dupptr, srcend);
             if (replicate < 3) {
-                uint8_t* nextsrc = srcptr + replicate;
+                uint8_t* nextsrc = dupptr + replicate;
                 uint8_t* nextdup = _packbits_next_replicate(nextsrc, srcend);
-                if (nextdup > nextsrc) {
-                    /* discard 2-byte run */
-                    dupptr = nextdup;
-                }
+                /* discard 2-byte run? */
+                dupptr = (nextdup > nextsrc) ? nextdup : dupptr;
             }
             literal = dupptr - srcptr;
         }
@@ -771,6 +778,9 @@ ssize_t imcd_packbits_encode(
             while (literal--) {
                 *dstptr++ = *srcptr++;
             }
+            /* memcpy(dstptr, srcptr, literal); */
+            /* dstptr += literal; */
+            /* srcptr += literal; */
         }
         else {
             dstptr = NULL;
@@ -779,8 +789,7 @@ ssize_t imcd_packbits_encode(
     }
 
     if (dstptr == NULL) {
-        if (dstsize < maxdst)
-        {
+        if (dstsize < maxdst) {
             return IMCD_OUTPUT_TOO_SMALL;
         }
         /* encoding exceeded maximum literal-only length */
@@ -788,13 +797,15 @@ ssize_t imcd_packbits_encode(
         literal = srcsize;
         srcptr = (uint8_t*)src;
         dstptr = dst;
-        while (srcptr < srcend)
-        {
+        while (srcptr < srcend) {
             literal = MIN_(128, srcend - srcptr);
             *dstptr++ = (uint8_t)(literal - 1);
             while (literal--) {
                 *dstptr++ = *srcptr++;
             }
+            /* memcpy(dstptr, srcptr, literal); */
+            /* dstptr += literal; */
+            /* srcptr += literal; */
         }
     }
     return (ssize_t)(dstptr - dst);
@@ -904,7 +915,7 @@ ssize_t imcd_packints_decode(
     const uint8_t* src,
     const ssize_t srcsize,  /** size of src in bytes */
     uint8_t* dst,  /** buffer to store unpacked items */
-    const ssize_t dstsize,   /** number of items to unpack */
+    const ssize_t items,  /** number of items to unpack */
     const int bps  /** number of bits in integer */
     )
 {
@@ -925,10 +936,10 @@ ssize_t imcd_packints_decode(
         case 16:
         case 32:
         case 64:
-            memcpy(dst, src, dstsize * itemsize);
-            return dstsize;
+            memcpy(dst, src, items * itemsize);
+            return items;
         case 1:
-            for (i = 0, j = 0; i < dstsize/8; i++) {
+            for (i = 0, j = 0; i < items/8; i++) {
                 value = src[i];
                 dst[j++] = (value & (uint8_t)(128)) >> 7;
                 dst[j++] = (value & (uint8_t)(64)) >> 6;
@@ -939,9 +950,9 @@ ssize_t imcd_packints_decode(
                 dst[j++] = (value & (uint8_t)(2)) >> 1;
                 dst[j++] = (value & (uint8_t)(1));
             }
-            if (dstsize % 8) {
+            if (items % 8) {
                 value = src[i];
-                switch (dstsize % 8)
+                switch (items % 8)
                 {
                     case 7: dst[j+6] = (value & (uint8_t)(2)) >> 1;
                     case 6: dst[j+5] = (value & (uint8_t)(4)) >> 2;
@@ -952,45 +963,45 @@ ssize_t imcd_packints_decode(
                     case 1: dst[j] = (value & (uint8_t)(128)) >> 7;
                 }
             }
-            return dstsize;
+            return items;
         case 2:
-            for (i = 0, j = 0; i < dstsize/4; i++) {
+            for (i = 0, j = 0; i < items/4; i++) {
                 value = src[i];
                 dst[j++] = (value & (uint8_t)(192)) >> 6;
                 dst[j++] = (value & (uint8_t)(48)) >> 4;
                 dst[j++] = (value & (uint8_t)(12)) >> 2;
                 dst[j++] = (value & (uint8_t)(3));
             }
-            if (dstsize % 4) {
+            if (items % 4) {
                 value = src[i];
-                switch (dstsize % 4)
+                switch (items % 4)
                 {
                     case 3: dst[j+2] = (value & (uint8_t)(12)) >> 2;
                     case 2: dst[j+1] = (value & (uint8_t)(48)) >> 4;
                     case 1: dst[j] = (value & (uint8_t)(192)) >> 6;
                 }
             }
-            return dstsize;
+            return items;
         case 4:
-            for (i = 0, j = 0; i < dstsize/2; i++) {
+            for (i = 0, j = 0; i < items/2; i++) {
                 value = src[i];
                 dst[j++] = (value & (uint8_t)(240)) >> 4;
                 dst[j++] = (value & (uint8_t)(15));
             }
-            if (dstsize % 2) {
+            if (items % 2) {
                 value = src[i];
                 dst[j] = (value & (uint8_t)(240)) >> 4;
             }
-            return dstsize;
+            return items;
         case 24:
             j = k = 0;
-            for (i = 0; i < dstsize; i++) {
+            for (i = 0; i < items; i++) {
                 dst[j++] = 0;
                 dst[j++] = src[k++];
                 dst[j++] = src[k++];
                 dst[j++] = src[k++];
             }
-            return dstsize;
+            return items;
     }
     /* 3, 5, 6, 7 */
     if (bps < 8) {
@@ -1001,7 +1012,7 @@ ssize_t imcd_packints_decode(
         val.b[IMCD_LSB] = src[j++];
         mask.b[IMCD_MSB] = imcd_bitmask(bps);
         mask.b[IMCD_LSB] = 0;
-        for (i = 0; i < dstsize; i++) {
+        for (i = 0; i < items; i++) {
             shr -= bps;
             tmp.i = (val.i & mask.i) >> shr;
             dst[k++] = tmp.b[IMCD_LSB];
@@ -1015,7 +1026,7 @@ ssize_t imcd_packints_decode(
                 mask.i >>= bps;
             }
         }
-        return dstsize;
+        return items;
     }
     /* 9, 10, 11, 12, 13, 14, 15 */
     if (bps < 16) {
@@ -1029,7 +1040,7 @@ ssize_t imcd_packints_decode(
         }
         mask.b[3] = 0xFF;
         mask.b[2] = imcd_bitmask(bps-8);
-        for (i = 0; i < dstsize; i++) {
+        for (i = 0; i < items; i++) {
             shr -= bps;
             tmp.i = (val.i & mask.i) >> shr;
             dst[k++] = tmp.b[0]; /* swap bytes */
@@ -1049,7 +1060,7 @@ ssize_t imcd_packints_decode(
 #else
     /* not implemented */
 #endif
-        return dstsize;
+        return items;
     }
     /* 17, 18, 19, 20, 21, 22, 23, 25, 26, 27, 28, 29, 30, 31 */
     if (bps < 32) {
@@ -1065,7 +1076,7 @@ ssize_t imcd_packints_decode(
         mask.b[6] = 0xFF;
         mask.b[5] = bps > 23 ? 0xFF : imcd_bitmask(bps - 16);
         mask.b[4] = bps < 24 ? 0x00 : imcd_bitmask(bps - 24);
-        for (i = 0; i < dstsize; i++) {
+        for (i = 0; i < items; i++) {
             shr -= bps;
             tmp.i = (val.i & mask.i) >> shr;
             dst[k++] = tmp.b[0]; /* swap bytes */
@@ -1091,7 +1102,7 @@ ssize_t imcd_packints_decode(
 #else
     /* Not implemented */
 #endif
-        return dstsize;
+        return items;
     }
     return IMCD_ERROR;
 }
@@ -1104,16 +1115,114 @@ ssize_t imcd_packints_encode(
     const uint8_t* src,
     const ssize_t srcsize,  /** size of src in bytes */
     uint8_t* dst,  /** buffer to store packed items */
-    const ssize_t dstsize,   /** number of items to pack */
+    const ssize_t items,  /** number of items to pack */
     const int bps  /** number of bits in integer */
     )
 {
+    ssize_t i, j, k;
+    ssize_t itemsize;
+    uint8_t value;
+
     if (srcsize == 0) {
         return 0;
     }
 
-    /* Input validation is done in wrapper function */
+    /* TODO: complete implementation */
 
+    /* Input validation is done in wrapper function */
+    itemsize = (ssize_t)(ceil(bps / 8.0));
+    itemsize = itemsize < 3 ? itemsize : itemsize > 4 ? 8 : 4;
+
+    switch (bps)
+    {
+        case 8:
+        case 16:
+        case 32:
+        case 64:
+            memcpy(dst, src, items * itemsize);
+            return items;
+        case 1:
+            for (i = 0, j = 0; j < items/8; j++) {
+                value = 0;
+                value |= (src[i++] << 7) & (uint8_t)(128);
+                value |= (src[i++] << 6) & (uint8_t)(64);
+                value |= (src[i++] << 5) & (uint8_t)(32);
+                value |= (src[i++] << 4) & (uint8_t)(16);
+                value |= (src[i++] << 3) & (uint8_t)(8);
+                value |= (src[i++] << 2) & (uint8_t)(4);
+                value |= (src[i++] << 1) & (uint8_t)(2);
+                value |= (src[i++] << 0) & (uint8_t)(1);
+                dst[j] = value;
+            }
+            if (items % 8) {
+                value = 0;
+                switch (items % 8)
+                {
+                    case 7: value |= (src[i++] << 7) & (uint8_t)(128);
+                    case 6: value |= (src[i++] << 6) & (uint8_t)(64);
+                    case 5: value |= (src[i++] << 5) & (uint8_t)(32);
+                    case 4: value |= (src[i++] << 4) & (uint8_t)(16);
+                    case 3: value |= (src[i++] << 3) & (uint8_t)(8);
+                    case 2: value |= (src[i++] << 2) & (uint8_t)(4);
+                    case 1: value |= (src[i++] << 1) & (uint8_t)(2);
+                }
+                dst[j++] = value;
+            }
+            return items;
+        case 2:
+            for (i = 0, j = 0; j < items/8; j++) {
+                value = 0;
+                value |= (src[i++] << 6) & (uint8_t)(192);
+                value |= (src[i++] << 4) & (uint8_t)(48);
+                value |= (src[i++] << 2) & (uint8_t)(12);
+                value |= (src[i++] << 0) & (uint8_t)(3);
+                dst[j] = value;
+            }
+            if (items % 8) {
+                value = 0;
+                switch (items % 8)
+                {
+                    case 3: value |= (src[i++] << 6) & (uint8_t)(192);
+                    case 2: value |= (src[i++] << 4) & (uint8_t)(48);
+                    case 1: value |= (src[i++] << 2) & (uint8_t)(12);
+                }
+                dst[j] = value;
+            }
+            return items;
+        case 4:
+            for (i = 0, j = 0; j < items/8; j++) {
+                value = 0;
+                value |= (src[i++] << 4) & (uint8_t)(240);
+                value |= (src[i++] << 0) & (uint8_t)(15);
+                dst[j] = value;
+            }
+            if (items % 8) {
+                dst[j] = (src[i++] << 4) & (uint8_t)(240);
+            }
+            return items;
+        // case 10:
+        //     return items;
+        case 12:
+            for (i = 0, j = 0; j < srcsize / 2; j+=3) {
+                value = 0;
+                value = src[i++] & (uint8_t)(15);
+                value = (src[i] >> 4) & (uint8_t)(15);
+                dst[j] = value;
+                value = 0;
+                value = (src[i++] << 4) & (uint8_t)(15);
+                value |= src[i++] & (uint8_t)(15);
+                dst[j+1] = value;
+                dst[j+2] = src[i++];
+            }
+            if (items % 2) {
+                dst[j] = (src[i++] << 4) & (uint8_t)(240);
+            }
+            return items;
+        // case 14:
+        //     return items;
+        // case 24:
+        //     return items;
+    }
     return IMCD_NOTIMPLEMENTED_ERROR;
 }
 
@@ -1604,17 +1713,23 @@ void imcd_lzw_del(imcd_lzw_handle_t* handle)
 {  \
     if ((bitcount + bitw) <= srcbitsize)  \
     {  \
-        const uint8_t* bytes = (uint8_t*)((void*)(src + (bitcount >> 3)));  \
-        code = (uint32_t) bytes[0];  \
-        code <<= 8;  \
-        code |= (uint32_t) bytes[1];  \
-        code <<= 8;  \
-        if ((bitcount + 24) <= srcbitsize)  \
-            code |= (uint32_t) bytes[2];  \
-        code <<= 8;  \
-        code <<= (uint32_t)(bitcount % 8);  \
-        code &= mask;  \
-        code >>= shr;  \
+        const uint32_t bitoffset = bitcount & 0x7;  \
+        const uint8_t* bytes = src + (bitcount >> 3);  \
+        if (bitoffset == 0 && bitw <= 24) {  \
+            code = (bytes[0] << 16) | (bytes[1] << 8) | bytes[2];  \
+            code >>= (24 - bitw);  \
+        } else {  \
+            code = (uint32_t) bytes[0];  \
+            code <<= 8;  \
+            code |= (uint32_t) bytes[1];  \
+            code <<= 8;  \
+            if ((bitcount + 24) <= srcbitsize)  \
+                code |= (uint32_t) bytes[2];  \
+            code <<= 8;  \
+            code <<= bitoffset;  \
+            code &= mask;  \
+            code >>= shr;  \
+        }  \
         bitcount += bitw;  \
     }  \
     else {code = LZW_EOI;}  \
@@ -1922,11 +2037,8 @@ ssize_t imcd_lzw_decode(
                 const uint8_t* oldbuffer = handle->buffer;
                 const ssize_t bufferlen = buffer - oldbuffer;
 
-                if (
-                    _lzw_alloc_buffer(handle, handle->buffersize - buffersize)
-                    <= 0
-                    )
-                {
+                if (_lzw_alloc_buffer(
+                        handle, handle->buffersize - buffersize) <= 0) {
                     return IMCD_MEMORY_ERROR;
                 }
                 if (handle->buffer != oldbuffer) {
@@ -1955,13 +2067,13 @@ ssize_t imcd_lzw_decode(
             else {
                 uint8_t* pstr = table[code].buf;
                 ssize_t len = table[code].len;
-                if (len > remaining) {
-                    len = remaining;
-                }
+                len = (len > remaining) ? remaining : len;
                 remaining -= len;
                 for (i = 0; i < len; i++) {
                     *dst++ = *pstr++;
                 }
+                /* memcpy(dst, table[code].buf, len); */
+                /* dst += len; */
             }
             /* table.append(table[oldcode] + table[code][0]) */
             table[tablesize].buf = buffer;
@@ -1973,13 +2085,11 @@ ssize_t imcd_lzw_decode(
                 for (i = 0; i < table[oldcode].len; i++) {
                     *buffer++ = *pstr++;
                 }
+                /* const ssize_t len = table[oldcode].len; */
+                /* memcpy(buffer, table[oldcode].buf, len); */
+                /* buffer += len; */
             }
-            if (code < 256) {
-                *buffer++ = (uint8_t) code;
-            }
-            else {
-                *buffer++ = table[code].buf[0];
-            }
+            *buffer++ = (code < 256) ? (uint8_t) code : table[code].buf[0];
         }
         else if (code > tablesize) {
             /* return dstsize - remaining; */
@@ -2000,13 +2110,13 @@ ssize_t imcd_lzw_decode(
             else {
                 uint8_t* pstr = table[oldcode].buf;
                 ssize_t len = table[oldcode].len;
-                if (len > remaining) {
-                    len = remaining;
-                }
+                len = (len > remaining) ? remaining : len;
                 remaining -= len;
                 for (i = 0; i < len; i++) {
                     *dst++ = *pstr++;
                 }
+                /* memcpy(dst, table[oldcode].buf, len); */
+                /* dst += len; */
                 if (--remaining < 0) break;
                 *dst++ = table[oldcode].buf[0];
             }
@@ -2134,7 +2244,7 @@ ssize_t imcd_lzw_encode(
 
         while (hash_keys[hashcode] >= 0) {
             if (hash_keys[hashcode] == hashkey) {
-                // Omega+K in table
+                /* Omega+K in table */
                 omega = hash_values[hashcode];
                 goto OUTER;
             }
@@ -2144,7 +2254,7 @@ ssize_t imcd_lzw_encode(
             }
         }
 
-        // Omega+K not in table
+        /* Omega+K not in table */
         /* add entry to table */
         hash_keys[hashcode] = hashkey;
         hash_values[hashcode] = nextcode++;
